@@ -2,9 +2,9 @@
 #
 #	XS.pm : perl function definition for Term::ReadLine::Gnu
 #
-#	$Id: XS.pm,v 1.8 1999-05-05 02:48:24+09 hayashi Exp $
+#	$Id: XS.pm,v 1.20 2002-07-27 22:39:49-05 hiroo Exp $
 #
-#	Copyright (c) 1999 Hiroo Hayashi.  All rights reserved.
+#	Copyright (c) 2002 Hiroo Hayashi.  All rights reserved.
 #
 #	This program is free software; you can redistribute it and/or
 #	modify it under the same terms as Perl itself.
@@ -36,6 +36,7 @@ use vars qw(*rl_unbind_function_in_map *rl_unbind_command_in_map);
 *rl_unbind_function_in_map = \&unbind_function;
 *rl_unbind_command_in_map  = \&unbind_command;
 
+rl_add_defun('history-expand-line',	 \&history_expand_line);
 # bind operate-and-get-next to \C-o by default for the compatibility
 # with bash and Term::ReadLine::Perl
 rl_add_defun('operate-and-get-next',	 \&operate_and_get_next, ord "\co");
@@ -53,6 +54,8 @@ my $rl_term_set = ',,,';
 # other purpose.
 my $_i;
 my @_matches;
+my @_tstrs;
+my $_tstrs_init = 0;
 
 1;
 
@@ -132,6 +135,30 @@ sub unbind_command ($;$) {
     }
 }
 
+sub rl_set_key ($$;$) {
+    my ($version) = $Attribs{library_version}
+	=~ /(\d+\.\d+)/;
+    if ($version < 4.2) {
+	carp "rl_set_key() is not supported.  Ignored\n";
+	return;
+    }
+    if (defined $_[2]) {
+	return _rl_set_key($_[0], _str2fn($_[1]), _str2map($_[2]));
+    } else {
+	return _rl_set_key($_[0], _str2fn($_[1]));
+    }
+}
+
+sub rl_macro_bind ($$;$) {
+    my ($version) = $Attribs{library_version}
+	=~ /(\d+\.\d+)/;
+    if (defined $_[2]) {
+	return _rl_macro_bind($_[0], $_[1], _str2map($_[2]));
+    } else {
+	return _rl_macro_bind($_[0], $_[1]);
+    }
+}
+
 sub rl_generic_bind ($$$;$) {
     if      ($_[0] == Term::ReadLine::Gnu::ISFUNC) {
 	if (defined $_[3]) {
@@ -155,7 +182,7 @@ sub rl_generic_bind ($$$;$) {
 	carp("Term::ReadLine::Gnu::rl_generic_bind: invalid \`type\'\n");
     }
 }
-	    
+
 sub rl_call_function ($;$$) {
     if (defined $_[2]) {
 	return _rl_call_function(_str2fn($_[0]), $_[1], $_[2]);
@@ -174,12 +201,41 @@ sub rl_invoking_keyseqs ($;$) {
     }
 }
 
+sub rl_add_funmap_entry ($$) {
+    my ($version) = $Attribs{library_version}
+	=~ /(\d+\.\d+)/;
+    if ($version < 4.2) {
+	carp "rl_add_funmap_entry() is not supported.  Ignored\n";
+	return;
+    }
+    return _rl_add_funmap_entry($_[0], _str2fn($_[1]));
+}
+
+sub rl_tty_set_default_bindings (;$) {
+    if (defined $_[0]) {
+	return _rl_tty_set_defaut_bindings(_str2map($_[1]));
+    } else {
+	return _rl_tty_set_defaut_bindings();
+    }
+}
+
 sub rl_message {
     my $fmt = shift;
     my $line = sprintf($fmt, @_);
     _rl_message($line);
 }
 
+sub rl_completion_mode {
+    # libreadline.* in Debian GNU/Linux 2.0 tells wrong value as '2.1-bash'
+    my ($version) = $Attribs{library_version}
+	=~ /(\d+\.\d+)/;
+    if ($version < 4.3) {
+	carp "rl_completion_mode() is not supported.  Ignored\n";
+	return;
+    }
+    return _rl_completion_mode(_str2fn($_[0]));
+}
+
 #
 #	for compatibility with Term::ReadLine::Perl
 #
@@ -187,8 +243,8 @@ sub rl_filename_list {
     my ($text) = @_;
 
     # lcd : lowest common denominator
-    my ($lcd, @matches) = completion_matches($text,
-					     \&filename_completion_function);
+    my ($lcd, @matches) = rl_completion_matches($text,
+						\&rl_filename_completion_function);
     return @matches ? @matches : $lcd;
 }
 
@@ -234,15 +290,27 @@ sub get_history_event ( $$;$ ) {
 # This module calls termcap (or its compatible) library, which the GNU
 # Readline Library already uses, instead of Term::Caps.pm.
 
+# Some terminals do not support 'ue' (underline end).
+use vars qw(%term_no_ue);
+%term_no_ue = ( kterm => 1 );
 
 sub ornaments {
     return $rl_term_set unless @_;
     $rl_term_set = shift;
     $rl_term_set ||= ',,,';
-    $rl_term_set = 'us,me,,' if $rl_term_set eq '1';
+    $rl_term_set = $term_no_ue{$ENV{TERM}} ? 'us,me,,' : 'us,ue,,'
+	if $rl_term_set eq '1';
     my @ts = split /,/, $rl_term_set, 4;
     my @rl_term_set
-	= map {$_ ? tgetstr($_) || '' : ''} @ts;
+	= map {
+	    # non-printing characters must be informed to readline
+	    my $t;
+	    ($_ and $t = tgetstr($_))
+		? (Term::ReadLine::Gnu::RL_PROMPT_START_IGNORE
+		   . $t
+		   . Term::ReadLine::Gnu::RL_PROMPT_END_IGNORE)
+		    : '';
+	} @ts;
     $Attribs{term_set} = \@rl_term_set;
     return $rl_term_set;
 }
@@ -251,6 +319,23 @@ sub ornaments {
 #	a sample custom function
 #
 
+# The equivalent of the Bash shell M-^ history-expand-line editing
+# command.
+
+# This routine was borrowed from bash.
+sub history_expand_line {
+    my ($count, $key) = @_;
+    my ($expanded, $new_line) = history_expand($Attribs{line_buffer});
+    if ($expanded > 0) {
+  	rl_modifying(0, $Attribs{end}); # save undo information
+  	$Attribs{line_buffer} = $new_line;
+    } elsif ($expanded < 0) {
+  	my $OUT = $Attribs{outstream};
+  	print $OUT "\n$new_line\n";
+  	rl_on_new_line();
+    }				# $expanded == 0 : no change
+}
+
 # The equivalent of the Korn shell C-o operate-and-get-next-history-line
 # editing command. 
 
@@ -313,7 +398,7 @@ sub change_ornaments {
     } elsif ($c =~ /v/i) {
 	ornaments('vb,,,');
     } else {
-	ding;
+	rl_ding;
     }
     rl_restore_prompt;
     rl_clear_message;
@@ -334,16 +419,35 @@ sub Tk_getc {
 #	$a->{redisplay_function} = $a->{shadow_redisplay};
 #	$line = $t->readline("password> ");
 sub shadow_redisplay {
+    @_tstrs = _tgetstrs() unless $_tstrs_init;
+    # remove prompt start/end mark from prompt string
+    my $prompt = $Attribs{prompt}; my $s;
+    $s = Term::ReadLine::Gnu::RL_PROMPT_START_IGNORE; $prompt =~ s/$s//g;
+    $s = Term::ReadLine::Gnu::RL_PROMPT_END_IGNORE;   $prompt =~ s/$s//g;
     my $OUT = $Attribs{outstream};
     my $oldfh = select($OUT); $| = 1; select($oldfh);
-    print $OUT (tgetstr('cr'), # carriage return
-		tgetstr('ce'), # clear to EOL
-		$Attribs{prompt}, '*' x length($Attribs{line_buffer}));
-    print $OUT (tgetstr('le') # cursor left
+    print $OUT ($_tstrs[0],	# carriage return
+		$_tstrs[1],	# clear to EOL
+		$prompt, '*' x length($Attribs{line_buffer}));
+    print $OUT ($_tstrs[2]	# cursor left
 		x (length($Attribs{line_buffer}) - $Attribs{point}));
     $oldfh = select($OUT); $| = 0; select($oldfh);
 }
 
+sub _tgetstrs {
+    my @s = (tgetstr('cr'),	# carriage return
+	     tgetstr('ce'),	# clear to EOL
+	     tgetstr('le'));	# cursor left
+    warn <<"EOM" unless (defined($s[0]) && defined($s[1]) && defined($s[2]));
+Your terminal 'TERM=$ENV{TERM}' does not support enough function.
+Check if your environment variable 'TERM' is set correctly.
+EOM
+    # suppress warning "Use of uninitialized value in print at ..."
+    $s[0] = $s[0] || ''; $s[1] = $s[1] || ''; $s[2] = $s[2] || '';
+    $_tstrs_init = 1;
+    return @s;
+}
+
 # callback handler wrapper function for CallbackHandlerInstall method
 sub _ch_wrapper {
     my $line = shift;
@@ -354,13 +458,13 @@ sub _ch_wrapper {
 	    ($result, $line) = history_expand($line);
 	    my $outstream = $Attribs{outstream};
 	    print $outstream "$line\n" if ($result);
-	    
+
 	    # return without adding line into history
 	    if ($result < 0 || $result == 2) {
 		return '';	# don't return `undef' which means EOF.
 	    }
 	}
-	
+
 	# add to history buffer
 	add_history($line) 
 	    if ($Attribs{MinLength} > 0
@@ -374,7 +478,7 @@ sub _ch_wrapper {
 #
 sub list_completion_function ( $$ ) {
     my($text, $state) = @_;
-    
+
     $_i = $state ? $_i + 1 : 0;	# clear counter at the first call
     my $cw = $Attribs{completion_word};
     for (; $_i <= $#{$cw}; $_i++) {
@@ -389,20 +493,23 @@ sub list_completion_function ( $$ ) {
 #
 sub _trp_completion_function ( $$ ) {
     my($text, $state) = @_;
-    
+
     my $cf;
     return undef unless defined ($cf = $Attribs{completion_function});
-    
+
     if ($state) {
 	$_i++;
     } else {
 	# the first call
 	$_i = 0;		# clear index
 	@_matches = &$cf($text,
-			$Attribs{line_buffer},
-			$Attribs{point} - length($text));
+			 $Attribs{line_buffer},
+			 $Attribs{point} - length($text));
+	# return here since $#_matches is 0 instead of -1 when
+	# @_matches = undef
+	return undef unless defined $_matches[0];
     }
-    
+
     for (; $_i <= $#_matches; $_i++) {
 	return $_matches[$_i] if ($_matches[$_i] =~ /^\Q$text/);
     }
